#!/opt/local/bin/python


import pygame, random, sys, math
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
from pygame.locals import *
from pprint import pprint
from math import *
from operator import *

class Vector:
    """quick drop-in hack for Scientific.Geometry.Vector"""
    def __init__(self,x,y,z):
        self.v = x,y,z
    
    def __add__(a,b):
        addtup = lambda tup: apply(add,tup)
        return apply(Vector,map(addtup,zip(a.v,b.v)))
    
    def __sub__(a,b):
        subtup = lambda tup: apply(sub,tup)
        return apply(Vector,map(subtup,zip(a.v,b.v)))
        
    def __getitem__(self,k):
        return self.v[k]
    
    def length(self):
        sq = lambda x: x*x
        return pow(sum(map(sq,self.v)),0.5)
        

# graphical helpers
def render_label(string):
    glRasterPos2f(0,0)
    for ch in string:
        glutBitmapCharacter(GLUT_BITMAP_8_BY_13,ord(ch))
def render_heading(string):
    glRasterPos2f(0,0)
    for ch in string:
        glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18,ord(ch))
        

class Branch:

    def __init__(self,id, src, dst):
        self.id = id # node identifier
        self.src = src # head
        self.dst = dst # tail (children grow here!)
        self.children = [] # subtrees
        self.weight = 1 # total nodes in subtrees including this one
        self.d = (dst-src).length() # our length, precomputed
    
    def render(self,d,t,intensity=0): # d is base distance, t is second tex coord
        sx,sy,sz = self.src
        dx,dy,dz = self.dst
        
        # line segment
        glBegin(GL_LINES)
        glTexCoord2f(t,d)
        glVertex3f(sx,sy,sz)
        glTexCoord2f(t,self.d+d)
        glVertex3f(dx,dy,dz)
        glEnd()
        glBegin(GL_POINTS)
        glVertex3f(dx,dy,dz)
        glEnd()
        
        # label
        if len(self.children):
            glPushAttrib(GL_CURRENT_BIT)
            glColor4f(1,1,1,intensity)
            glPushMatrix()
            glTranslatef(dx,dy,dz)
            render_label(str(self.id))
            glPopMatrix()
            glPopAttrib()
        
        # children
        for child in self.children:
            child.render(d+self.d,t,intensity)
    
    def grow(self,id):
        r = random.randint(0,self.weight-1)
        self.weight += 1 # we are adding a child SOMEWHERE no matter what
        
        if r == 0:
            # add the child to this branch
            
            sx,sy,sz = self.src
            dx,dy,dz = self.dst
            
            ruffle = 0.5*self.d
            momentum = 0.75
            c_dx = dx + (dx-sx)*momentum + 2.0*ruffle*(random.random()-0.5)
            c_dy = dy + (dy-sy)*momentum + 2.0*ruffle*(random.random()-0.5)
            c_dz = dz + (dz-sz)*momentum + 2.0*ruffle*(random.random()-0.5)
            
            c_src = self.dst
            c_dst = Vector(c_dx,c_dy,c_dz)
            
            child = Branch(id,c_src, c_dst)
            
            self.children.append(child)
        
        else:
            # delegating growth  to a child
            nch = len(self.children)
            if nch == 1:
                i = 0
            else:
                i = random.randint(0,nch-1)
            self.children[i].grow(id) 
            
    # end method grow
    
    def prune(self, id,onprune=None):
        """prune a child branch by id, signaling parent when found"""
        if id == self.id:
            return True
        else:
            # check for child to remove
            did_prune = False
            for ch in self.children:
                if ch.prune(id,onprune):
                    did_prune = True
                    self.children.remove(ch)
            # recalc weight
            if did_prune:
                self.weight = 1
                if self.children:
                    self.weight += 1 + sum([ch.weight for ch in self.children])
                if onprune: onprune()
                return False
    # end method prune

    def score(self):
        """calculate farthest child node's distance from the origin"""
        if self.children:
            return max([ch.score() for ch in self.children])
        else:
            return self.dst.length()
            
    
class BushGame:

    def __init__(self,*args):
        """defines guts of a BushGame container object"""
        # misc
        self.args = args
        self.last_ticks = 0
        self.ticks = 0
        self.dt = 0
        self.running = True
        self.return_code = 0
        
        # rules
        self.text_box = ""
        self.bush = None
        self.last_grow = 0
        self.grow_delay = 4.0/3.0 # match fralb tempo
        
        self.last_id = None
        self.max_points = 0
        self.lastscore = 1
        
        self.id_grow = 1
        self.id_prune = 2
        
        # io
        self.win_width = 1024
        self.win_height = 768
        self.show_fullscreen = False
        
        # graphics
        self.tex_pat = None
        self.tex_welcome = None
        
        self.intensity = 1
        self.line_shift2 = 0
        self.line_shift = 0
        self.line_speed2 = 0.2
        self.line_speed = -0.2
        
        # sound
        pass
        
        
        # initialization logic
        self.init_io()
        self.init_gfx()
        self.init_snd()
        self.init_rules()
        
        # all done, call run() to play!
        
    # end method __init__
    
    def init_io(self):
        """pygame input/output setup"""
        pygame.mixer.pre_init(44100) # "CD Quality" audio
        pygame.init()
        
        flags = OPENGL|DOUBLEBUF
        if self.show_fullscreen:
            flags |= FULLSCREEN
        pygame.display.set_mode((self.win_width,self.win_height), flags)
        pygame.key.set_repeat(100,30)
        self.ticks = pygame.time.get_ticks()
        self.last_ticks = self.ticks-1
        
    # end method init_io

        
    def init_gfx(self):
        """gl setup and graphical asset loading"""
        
        w, h = self.win_width, self.win_height
        glViewport(0, 0, w,h)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(60, 1.0*w/h, 0.1, 100.0)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        
        glEnable(GL_TEXTURE_2D)
        glShadeModel(GL_SMOOTH)
        glClearColor(0,0,0,0)
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST)
        glBlendFunc( GL_SRC_ALPHA, GL_ONE )
        glEnable(GL_BLEND)
        glDisable(GL_DEPTH_TEST)
        glEnable(GL_LINE_SMOOTH)
        glEnable(GL_POINT_SMOOTH)
        

        # texturing 
        (self.tex_pat,self.tex_welcome) = glGenTextures(2)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
    
        # tex_pat
        
        glBindTexture(GL_TEXTURE_2D, self.tex_pat)
        surf = pygame.image.load('starbles.png')
        data = pygame.image.tostring(surf, "RGBA", 1)
        gluBuild2DMipmaps(GL_TEXTURE_2D,4,surf.get_width(), surf.get_height(), GL_RGBA,GL_UNSIGNED_BYTE,data)
        
        
        # line params
        glLineWidth(3.0)
        glPointSize(8.0)
              

    # end method init_gfx   

    
    def init_snd(self):
        """music and sound setup"""
        pygame.mixer.music.load("fralbs.ogg")
        pygame.mixer.music.play(-1)
        
        self.snd_prune = pygame.mixer.Sound("skroik.ogg")
        self.snd_type = pygame.mixer.Sound("beep.ogg")
        self.snd_neg = pygame.mixer.Sound("neg.ogg")
        
        # TODO: load additional sounds
        pass
        
    def init_rules(self):
        
        # generate stump
        src = Vector(0,0,0)
        dst = Vector(0,0,1)
        self.last_id = 0
        self.stump = Branch(self.last_id,src,dst)
        self.last_id += 1
        
        self.recalc_score()
        
    def run(self):
        while self.running:
            # handle all events until NOEVENT
            while 1: 
                event = pygame.event.poll()
                if event.type == NOEVENT:
                    break
                if event.type in [KEYDOWN,KEYUP]:
                    self.handle_KeyboardEvent(event)
                if event.type == USEREVENT:
                    self.handle_UserEvent(event)
                if event.type == QUIT:
                    self.quit()
            
            self.calc_dt()
            
            self.tick_rules()
            self.tick_gfx()
            self.render()
            
        # running is now false and return code was updated from quit
        return self.return_code
            
        
    def handle_KeyboardEvent(self,event):
        k = event.key
        if k==K_ESCAPE:
            self.quit()
            
        # virtual text box input
        if event.type == KEYDOWN:
            if k >= K_0 and k <= K_9:
                if k != K_0 or len(self.text_box):                    
                    self.text_box += chr(k)
                    self.snd_type.play()
            elif k == K_BACKSPACE:
                if len(self.text_box):
                    self.text_box = self.text_box[0:-1]
                    self.snd_type.play()
            elif k == K_RETURN:
                try:
                    number = int(self.text_box)
                    e = pygame.event.Event(USEREVENT,code=self.id_prune,id=number)
                    pygame.event.post(e)
                except:
                    pass
                self.text_box = ""
            elif k==K_SPACE:
                self.intensity = (self.intensity+0.25)%1.0
        
    def handle_UserEvent(self,event):
        """turn game events into game state changes"""
        if event.code == self.id_prune:
            if event.id != 0:
                self.stump.prune(event.id,onprune=lambda: self.snd_prune.play())
        elif event.code == self.id_grow:
            self.last_id += 1
            self.stump.grow(self.last_id)
        self.recalc_score()
            
    # end method handle_UserEvent
    
    def recalc_score(self):
        score = self.stump.score()
        weight = self.stump.weight
        points = int(100*weight/score)
        self.max_points = max(points,self.max_points)
        self.heading = "allure %d/%d" % (points, self.max_points)
        self.ratio = (1.0*points/self.max_points)**4.0
        if score > self.lastscore:
            self.snd_neg.play()
        self.lastscore = score
        
    
    def calc_dt(self):
        """update game time"""
        self.last_ticks = self.ticks
        self.ticks = pygame.time.get_ticks()/1000.0
        self.dt = self.ticks - self.last_ticks
        # XXX: timehack
        self.dt = 0.0001
    
    def tick_physics(self):
        self.world.step(self.dt)
        
    def tick_rules(self):
        # check if we should grow again
        if self.T() > self.last_grow + self.grow_delay:
            self.last_grow = self.T()
            e = pygame.event.Event(USEREVENT,code=self.id_grow)
            pygame.event.post(e)

        
    def tick_gfx(self):
        t = 1.0/self.line_speed
        self.line_shift = (self.T()%t)/t
        t2 = 1.0/self.line_speed2
        self.line_shift2 = (self.T()%t2)/t2
        pass
        
    def render(self):
        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT )
        glBindTexture( GL_TEXTURE_2D, self.tex_pat )
        glLoadIdentity()
        
        global sc
        sc += 0.01*(self.stump.score()-sc)
        gluLookAt(0,2+(1+sin(self.T()/3.0))*1*sc,0.5+(1+sin(self.T()/6.0))*sc/4, 0,0,1.0+(1+sin(self.T()/8.0))*sc/2, 0,0,1)
        glRotatef(self.T()*20.0,0,0,1)
         
        glColor4f(1,1,1,1)
        if self.text_box: 
            render_label("self.stump.prune(%s)?" % self.text_box)
        
        glTexCoord2f(0,0)
        glutSolidCube(0.25)
        glBegin(GL_QUADS)
        glVertex3f(-5,-5,-0.25)
        glVertex3f(-5,+5,-0.25)
        glVertex3f(+5,+5,-0.25)
        glVertex3f(+5,-5,-0.25)
        glVertex3f(-1.25,-1.25,-0.0)
        glVertex3f(-1.25,+1.25,-0.0)
        glVertex3f(+1.25,+1.25,-0.0)
        glVertex3f(+1.25,-1.25,-0.0)
        glEnd()
        self.stump.render(self.line_shift, self.line_shift2,self.intensity)
        
        glTranslatef(0,0,+0.2)
        glColor4f(1,self.ratio,self.ratio,1) 
        render_heading(self.heading)
                
        # show the user!
        pygame.display.flip()
        
    # end method render
    
    def quit(self, code=0):
        self.running = False
        self.return_code = code
    
    # utility functions (might not ever be called)    
    def ar(self): return 1.0*self.win_width/self.win_height
    def T(self): return self.ticks
    def FPS(self): return 1.0/self.dt 
    def SPF(self): return self.dt/1.0
    def r(self, vmax=1.0,vmin=0.0): return random.uniform(vmin,vmax)
    def g(self, mu=0,sigma=1): return random.gauss(mu,sigma)
    
    def buildTransform(self,o):
        x,y,z = o.getPosition()
        rot = o.getRotation()
        return( rot[0],rot[3],rot[6], 0,
                rot[1],rot[4],rot[7], 0,
                rot[2],rot[5],rot[8], 0,
                x,y,z,1.0 )
                
# end class BushGame
sc = 1

# driver
if __name__ == '__main__':
    g = BushGame(sys.argv)
    ret = g.run()
    sys.exit(ret)
# end driver
